from typing import List, Tuple, Any
import os
import shutil
import logging
import time
from openai import OpenAIError, RateLimitError, APIConnectionError

from langchain_community.vectorstores import FAISS as VectorStore
from langchain_community.retrievers import BM25Retriever
from langchain_openai import OpenAIEmbeddings, ChatOpenAI
from langchain.chains import RetrievalQA
from langchain.schema import Document

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def perform_bm25_retrieval(
    docs: List[Document], split_docs: List[Document], query: str, num_docs: int
) -> List[Document]:
    """
    Retrieve documents using BM25 keyword matching.
    """
    try:
        all_docs = docs + split_docs
        if not all_docs:
            raise ValueError("No documents provided for BM25 retrieval")
        
        logger.info("Retrieval using Keyword Matching Algorithm.")
        bm25_retriever = BM25Retriever.from_documents(all_docs)
        bm25_retriever.k = 2 * num_docs if num_docs < 10 else num_docs + 10
        
        retrieved_docs = bm25_retriever.invoke(query)
        logger.info(f"BM25 retrieval completed. Retrieved {len(retrieved_docs)} documents.")
        return retrieved_docs
        
    except Exception as e:
        logger.error(f"BM25 retrieval failed: {str(e)}")
        # Return all documents as fallback
        logger.warning("Falling back to returning all documents")
        return docs + split_docs


def perform_dense_retrieval(
    retrieved_docs: List[Document],
    key_docs: List[Document],
    chroma_db_path: str,
    openai_api_key: str,
    num_docs: int,
    embedding_model: str = "text-embedding-3-small"
) -> Tuple[Any, Any]:
    """
    Retrieve documents using dense semantic search (FAISS + OpenAI embeddings).
    """
    try:
        all_docs = retrieved_docs + key_docs
        if not all_docs:
            raise ValueError("No documents provided for dense retrieval")
        
        # Validate document types
        for doc in all_docs:
            if not isinstance(doc, Document):
                raise ValueError("The retrieved documents must be of type Document.")
        
        logger.info("Retrieval using FAISS semantic search.")
        
        # Create embeddings with retry logic
        embedding = _create_embeddings_with_retry(embedding_model, openai_api_key)
        
        # Clean up existing database
        if os.path.exists(chroma_db_path):
            try:
                shutil.rmtree(chroma_db_path)
            except OSError as e:
                logger.warning(f"Could not remove existing chroma_db: {e}")
        
        # Create vector database
        vector_db = VectorStore.from_documents(all_docs, embedding)
        dense_retriever = vector_db.as_retriever(search_kwargs={'k': num_docs})
        
        logger.info(f"Dense retrieval setup completed with {len(all_docs)} documents.")
        return dense_retriever, vector_db
        
    except OpenAIError as e:
        logger.error(f"OpenAI API error during dense retrieval: {str(e)}")
        raise
    except Exception as e:
        logger.error(f"Dense retrieval failed: {str(e)}")
        raise


def _create_embeddings_with_retry(embedding_model: str, openai_api_key: str, max_retries: int = 3) -> OpenAIEmbeddings:
    """
    Create OpenAI embeddings with retry logic for rate limiting and connection errors.
    """
    last_exception = None
    
    for attempt in range(max_retries):
        try:
            embedding = OpenAIEmbeddings(model=embedding_model, openai_api_key=openai_api_key)
            return embedding
        except RateLimitError as e:
            last_exception = e
            if attempt < max_retries - 1:
                wait_time = 2 ** attempt  # Exponential backoff
                logger.warning(f"Rate limit hit, waiting {wait_time} seconds before retry {attempt + 1}/{max_retries}")
                time.sleep(wait_time)
            else:
                logger.error(f"Failed to create embeddings after {max_retries} attempts due to rate limiting")
                raise
        except APIConnectionError as e:
            last_exception = e
            if attempt < max_retries - 1:
                wait_time = 2 ** attempt
                logger.warning(f"API connection error, waiting {wait_time} seconds before retry {attempt + 1}/{max_retries}")
                time.sleep(wait_time)
            else:
                logger.error(f"Failed to create embeddings after {max_retries} attempts due to connection issues")
                raise
        except Exception as e:
            logger.error(f"Unexpected error creating embeddings: {str(e)}")
            raise
    
    # This should never be reached due to the raise statements above, but added for type safety
    if last_exception:
        raise last_exception
    raise RuntimeError(f"Failed to create embeddings after {max_retries} attempts")


def initialize_llm_and_qa_chain(
    dense_retriever: Any, openai_api_key: str, model: str
) -> Tuple[Any, Any]:
    """
    Initialize the LLM and QA chain.
    """
    try:
        logger.info(f"Initializing LLM with model: {model}")
        llm = ChatOpenAI(model=model, openai_api_key=openai_api_key)
        qa_chain = RetrievalQA.from_chain_type(
            llm=llm,
            chain_type="stuff",
            retriever=dense_retriever,
            return_source_documents=True
        )
        logger.info("LLM and QA chain initialized successfully")
        return llm, qa_chain
    except Exception as e:
        logger.error(f"Failed to initialize LLM and QA chain: {str(e)}")
        raise


def print_results(response: dict) -> None:
    """
    Print the RAG answer and metadata of retrieved sources.
    """
    print("\nRAG Answer:\n", response["result"])
    print("\nNumber of Retrieved Sources:", len(response["source_documents"]))
    print("\nRetrieved Sources Metadata:")
    for i, doc in enumerate(response["source_documents"], start=1):
        source = doc.metadata.get("source", "Unknown source")
        distributors = doc.metadata.get("distributors")
        interventions = doc.metadata.get("interventions")
        chunk_id = doc.metadata.get("chunk_id")
        chunk_size = doc.metadata.get("chunk_size")
        chunk_head = doc.metadata.get("chunk_head")
        print(f"Text Chunk {i}:")
        print(f"   - source: {source}")
        if chunk_id is not None:
            print(f"   - chunk_id: {chunk_id}")
        if chunk_size is not None:
            print(f"   - chunk_size: {chunk_size}")
        if chunk_head is not None:
            print(f"   - chunk_head: {chunk_head}")
        if interventions is not None:
            print(f"   - interventions: {interventions}")
        if distributors is not None:
            print(f"   - distributors: {distributors}")


def reset_chroma_db(chroma_db_fusion: Any) -> None:
    """
    Reset the Chroma database.
    """
    from chromadb.api.client import SharedSystemClient
    SharedSystemClient.clear_system_cache()
    all_ids = chroma_db_fusion.get()["ids"]
    chroma_db_fusion._collection.delete(ids=all_ids)
    chroma_db_fusion.delete_collection()


def generate_retriever(
    chroma_db_path: str,
    docs: List[Document],
    key_docs: List[Document],
    num_docs: int,
    openai_api_key: str,
    query: str,
    split_docs: List[Document]
) -> Tuple[Any, Any]:
    """
    Generate a retriever using BM25 and dense retrieval.
    """
    if not openai_api_key:
        raise ValueError("OpenAI API key is required.")
    if not docs:
        raise ValueError("No documents provided for retrieval.")
    retrieved_docs = perform_bm25_retrieval(docs, split_docs, query, num_docs)
    dense_retriever, vector_db = perform_dense_retrieval(
        retrieved_docs, key_docs, chroma_db_path, openai_api_key, num_docs
    )
    return vector_db, dense_retriever


def ask_llm(
    openai_api_key: str,
    query: str,
    dense_retriever: Any,
    vector_db: Any,
    model: str,
    max_retries: int = 3
) -> Tuple[dict, Any, Any]:
    """
    Query the LLM using the QA chain and print results with retry logic.
    """
    if not openai_api_key:
        raise ValueError("OpenAI API key is required. Please provide it in the environment variable 'OPENAI_API_KEY' or "
                         "set it in the 'key.env' file as: OPENAI_API_KEY='your_openai_key_here'.")
    
    try:
        llm, qa_chain = initialize_llm_and_qa_chain(dense_retriever, openai_api_key, model=model)
        
        # Retry logic for LLM query
        for attempt in range(max_retries):
            try:
                logger.info(f"Querying LLM (attempt {attempt + 1}/{max_retries})")
                response = qa_chain.invoke({"query": query})
                print_results(response)
                logger.info("LLM query completed successfully")
                break
            except RateLimitError as e:
                if attempt < max_retries - 1:
                    wait_time = 2 ** attempt
                    logger.warning(f"Rate limit hit during LLM query, waiting {wait_time} seconds before retry")
                    time.sleep(wait_time)
                else:
                    logger.error(f"LLM query failed after {max_retries} attempts due to rate limiting: {str(e)}")
                    raise
            except APIConnectionError as e:
                if attempt < max_retries - 1:
                    wait_time = 2 ** attempt
                    logger.warning(f"API connection error during LLM query, waiting {wait_time} seconds before retry")
                    time.sleep(wait_time)
                else:
                    logger.error(f"LLM query failed after {max_retries} attempts due to connection issues: {str(e)}")
                    raise
            except OpenAIError as e:
                logger.error(f"OpenAI API error during LLM query: {str(e)}")
                raise
            except Exception as e:
                if attempt < max_retries - 1:
                    logger.warning(f"Unexpected error during LLM query, retrying: {str(e)}")
                    time.sleep(1)
                else:
                    logger.error(f"LLM query failed after {max_retries} attempts: {str(e)}")
                    raise
        
        # Cleanup vector database
        try:
            if not isinstance(vector_db, VectorStore):
                reset_chroma_db(vector_db)
        except Exception as e:
            logger.warning(f"Could not reset chroma database: {str(e)}")
        
        return response, llm, qa_chain
        
    except Exception as e:
        logger.error(f"Critical error in ask_llm: {str(e)}")
        raise

