import streamlit as st
import tempfile
import os
import logging
from typing import List
from langchain_community.document_loaders import (
    JSONLoader, PythonLoader, PyPDFLoader, TextLoader, UnstructuredWordDocumentLoader
)
from langchain_text_splitters import RecursiveCharacterTextSplitter, Language
from langchain.schema import Document
from emodpy_converter.template import extract_and_save_code, load_query
from emodpy_converter.doc_loader import json_metadata_func, _load_html
from emodpy_converter.ask_llm import ask_llm, generate_retriever
from emodpy_converter.secrets_loader import get_openai_api_key
from openai import OpenAIError

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


# --- UI Setup ---
st.title("🔗EMOD Campaign JSON to Emodpy Script Converter.")

openai_api_key = st.sidebar.text_input("Please enter your OpenAI API key", type="password")
#if not openai_api_key:
#    st.sidebar.warning("Please enter your OpenAI API key to use this tool.")
openai_api_key = get_openai_api_key(openai_api_key)
if not openai_api_key:
    st.sidebar.warning("Please enter your OpenAI API key to use this tool.")
models = ["o3-mini-2025-01-31", "gpt-4o-2024-08-06", "gpt-4.1-2025-04-14"]
model = st.sidebar.selectbox("Select the model to use", models, index=0)
st.sidebar.write('You selected model:', model)

st.markdown("Upload campaign json file (json) and internal documents (PDF, HTML or Python script) to analyze using AI.")

# --- File Uploads ---
json_file = st.file_uploader("Upload json files that you want to convert to Emodpy script", type=["json"])
doc_files = st.file_uploader("Upload documents (PDF, HTML) and example python script (.py)", type=["pdf", "html", "py"], accept_multiple_files=True)
key_doc_files = st.file_uploader("Optional: Upload key documents (PDF, HTML) and example python script (.py); these files will not be filtered out by RAG.", type=["pdf", "html", "py"], accept_multiple_files=True)
split_doc_files = st.file_uploader("Optional: Upload optional documents (PDF, HTML) and example python script (.py); these files will be split and filtered by RAG.", type=["pdf", "html", "py"], accept_multiple_files=True)
optional_query = st.text_input("Optional: Enter additional keywords to help AI understand the context of the json file. For example, 'ControlledVaccine'")
submit = st.button("Generate Emodpy script")


# --- Document Processing ---
def load_document(file, suffix):
    """Load a document with comprehensive error handling and cleanup."""
    tmp_file_path = None
    try:
        with tempfile.NamedTemporaryFile(delete=False, suffix="." + suffix) as tmp_file:
            tmp_file_path = tmp_file.name
            file_content = file.read()
            
            if not file_content:
                logger.warning(f"File {file.name} appears to be empty")
                return []
            
            tmp_file.write(file_content)
            tmp_file.flush()
            
        # Load document based on file type
        if suffix == "pdf":
            loader = PyPDFLoader(tmp_file_path)
            docs = loader.load()
        elif suffix == "html":
            from pathlib import Path
            docs = [_load_html(Path(tmp_file_path))]
        elif suffix == "py":
            loader = PythonLoader(tmp_file_path)
            docs = loader.load()
        elif suffix == "json":
            loader = JSONLoader(
                tmp_file_path, 
                jq_schema='.', 
                content_key="Events", 
                text_content=False, 
                metadata_func=json_metadata_func
            )
            docs = loader.load()
        elif suffix == "txt":
            loader = TextLoader(tmp_file_path)
            docs = loader.load()
        elif suffix == "docx":
            loader = UnstructuredWordDocumentLoader(tmp_file_path)
            docs = loader.load()
        else:
            logger.warning(f"Unsupported file type: {suffix}")
            return []
        
        if not docs:
            logger.warning(f"No content extracted from {file.name}")
            return []
        
        logger.info(f"Successfully loaded {len(docs)} document(s) from {file.name}")
        return docs
        
    except Exception as e:
        logger.error(f"Error loading document {file.name}: {str(e)}")
        st.error(f"❌ Failed to process {file.name}: {str(e)}")
        return []
        
    finally:
        # Cleanup temporary file
        if tmp_file_path and os.path.exists(tmp_file_path):
            try:
                os.unlink(tmp_file_path)
            except OSError as e:
                logger.warning(f"Could not remove temporary file {tmp_file_path}: {e}")


def process_documents(files) -> List[Document]:
    """Process multiple documents with improved error handling and progress tracking."""
    docs = []
    if not files:
        return docs
    
    # Create progress bar for processing multiple files
    progress_bar = st.progress(0)
    status_text = st.empty()
    
    successful_files = 0
    failed_files = 0
    
    for i, f in enumerate(files):
        try:
            status_text.text(f"Processing {f.name}...")
            progress_bar.progress((i + 1) / len(files))
            
            # Validate file extension
            if "." not in f.name:
                st.warning(f"⚠️ File {f.name} has no extension, skipping")
                failed_files += 1
                continue
            
            suffix = f.name.split(".")[-1].lower()
            
            # Check file size (limit to 50MB)
            if hasattr(f, 'size') and f.size > 50 * 1024 * 1024:
                st.warning(f"⚠️ File {f.name} is too large (>50MB), skipping")
                failed_files += 1
                continue
            
            file_docs = load_document(f, suffix)
            if file_docs:
                docs.extend(file_docs)
                successful_files += 1
            else:
                failed_files += 1
                
        except Exception as e:
            logger.error(f"Unexpected error processing {f.name}: {str(e)}")
            st.error(f"❌ Unexpected error processing {f.name}: {str(e)}")
            failed_files += 1
    
    # Clear progress indicators
    progress_bar.empty()
    status_text.empty()
    
    # Show summary
    if successful_files > 0:
        st.success(f"✅ Successfully processed {successful_files} file(s)")
    if failed_files > 0:
        st.warning(f"⚠️ Failed to process {failed_files} file(s)")
    
    logger.info(f"Document processing complete. Success: {successful_files}, Failed: {failed_files}")
    return docs


def split_documents(docs) -> List[Document]:
    """Split documents into chunks with error handling."""
    try:
        if not docs:
            logger.warning("No documents provided for splitting")
            return []
        
        splitter = RecursiveCharacterTextSplitter.from_language(
            language=Language.PYTHON,
            chunk_size=10000,
            chunk_overlap=200
        )
        
        split_docs = splitter.split_documents(docs)
        
        # Add metadata to chunks
        for i, doc in enumerate(split_docs):
            doc.metadata["chunk_id"] = i + 1
            doc.metadata["chunk_size"] = len(doc.page_content)
            doc.metadata["chunk_head"] = doc.page_content[:100].rstrip()
        
        logger.info(f"Successfully split {len(docs)} documents into {len(split_docs)} chunks")
        return split_docs
        
    except Exception as e:
        logger.error(f"Error splitting documents: {str(e)}")
        st.error(f"❌ Error splitting documents: {str(e)}")
        # Return original documents as fallback
        return docs


# --- Query Preparation ---
query = None
if json_file:
    try:
        if isinstance(json_file, list):
            st.warning("Multiple json files detected. Using the first one.")
            json_file = json_file[0]
        
        query = load_query(json_file)
        if optional_query:
            query += f"\n\nKeywords: {optional_query}"
        
        logger.info("Query prepared successfully")
        
    except Exception as e:
        logger.error(f"Error preparing query from JSON file: {str(e)}")
        st.error(f"❌ Error processing JSON file: {str(e)}")
        st.stop()
else:
    st.warning("Please upload a json file to convert to Emodpy script.")

if not doc_files:
    st.warning("Please upload at least one document (PDF, HTML, or Python script) to analyze.")

if not openai_api_key:
    st.stop()
# --- Main Processing ---
if submit:
    try:
        # Initialize processing status
        processing_status = st.empty()
        processing_status.info("🔄 Starting document processing...")
        
        # Process documents with error handling
        raw_docs = process_documents(doc_files) if doc_files else []
        if doc_files and not raw_docs:
            st.error("❌ No valid documents could be processed. Please check your file formats and try again.")
            st.stop()
        
        key_docs = process_documents(key_doc_files) if key_doc_files else []
        
        split_docs = process_documents(split_doc_files) if split_doc_files else []
        if split_doc_files and split_docs:
            split_docs = split_documents(split_docs)
        elif split_doc_files:
            st.warning("No valid split documents could be processed.")

        # Display results in order
        if doc_files:
            st.subheader("📄 Docs")
            st.write(f"Processed {len(raw_docs)} document(s)")

        if key_doc_files:
            st.subheader("📄 Key Docs")
            st.write(f"Processed {len(key_docs)} key document(s)")

        if split_doc_files and split_docs:
            st.subheader("📄 Split Docs")
            st.write(f"Created {len(split_docs)} document chunks")

        # Validate minimum requirements
        if not json_file:
            st.error("❌ JSON file is required")
            st.stop()
        
        if not openai_api_key:
            st.error("❌ OpenAI API key is required")
            st.stop()
        
        if not raw_docs:
            st.error("❌ At least one document must be successfully processed")
            st.stop()

        # Main AI processing
        st.subheader("📊 AI Response")
        processing_status.info("🤖 Generating retriever and querying AI...")
        
        with st.spinner("Ask LLM to convert json to Emodpy script..."):
            try:
                chroma_db_path = './chroma_db'
                num_docs = 12
                
                # Generate retriever with error handling
                chroma_db_fusion, dense_retriever = generate_retriever(
                    chroma_db_path, raw_docs, key_docs, num_docs, openai_api_key, query, split_docs
                )
                
                # Query LLM with error handling
                response, llm, qa_chain = ask_llm(
                    openai_api_key, query, dense_retriever, chroma_db_fusion, model=model
                )
                
                # Save generated code
                output_folder = "./output"
                os.makedirs(output_folder, exist_ok=True)
                output_filename = json_file.name.replace(".json", ".py")
                extract_and_save_code(response["result"], os.path.join(output_folder, output_filename))
                
                processing_status.success("✅ Processing completed successfully!")
                
            except OpenAIError as e:
                logger.error(f"OpenAI API error: {str(e)}")
                st.error(f"❌ OpenAI API Error: {str(e)}")
                st.error("Please check your API key and try again. If the error persists, you may have hit rate limits.")
                st.stop()
                
            except Exception as e:
                logger.error(f"Error during AI processing: {str(e)}")
                st.error(f"❌ AI Processing Error: {str(e)}")
                st.error("An unexpected error occurred during AI processing. Please try again.")
                st.stop()

        # Display results
        if 'response' in locals():
            st.write(response["result"])
            st.subheader("Original Json")
            try:
                json_content = json_file.getvalue().decode("utf-8")
                st.text_area("Original Json File Content", json_content, height=300)
            except Exception as e:
                logger.error(f"Error displaying JSON content: {str(e)}")
                st.error("❌ Could not display original JSON content")
        
    except Exception as e:
        logger.error(f"Critical error in main processing: {str(e)}")
        st.error(f"❌ Critical Error: {str(e)}")
        st.error("A critical error occurred. Please refresh the page and try again.")


